@@ -1,20 +1,20 @@
 
+from collections import OrderedDict
 from datetime import datetime
 
-from bson import SON
-from bson.py3compat import string_type
-from bson.decimal128 import Decimal128
-from bson.timestamp import Timestamp
-
 from ..errors import WriteError
-from .core import (
-    _cmp_decimal,
-    FieldWalker,
-    Weighted,
-    FieldWriteError,
+
+from .field_walker import FieldWalker, FieldWriteError
+from .weighted import Weighted, _cmp_decimal
+from .queries import QueryFilter, ordering
+from ..types import (
+    Decimal128,
+    Timestamp,
+    string_types,
+    is_numeric_type,
+    is_duckument_type,
+    is_integer_type,
 )
-from .queries import QueryFilter
-from .helpers import is_numeric_type, is_duckument_type
 
 
 def _update(fieldwalker,
@@ -67,17 +67,17 @@ class Updator(object):
             "$pull": parse_pull,
             "$push": parse_push,
             "$pullAll": parse_pull_all,
-            "$each": None,
-            "$position": None,
-            "$slice": None,
-            "$sort": None,
 
+            # $each             implemented in Eacher
+            # $position         implemented in Eacher
+            # $slice            implemented in Eacher
+            # $sort             implemented in Eacher
         }
 
         self.fields_to_update = []
         self.array_filters = self.array_filter_parser(array_filters or [])
         # sort by key (operator)
-        self.operations = SON(sorted(self.parser(spec).items()))
+        self.operations = OrderedDict(sorted(self.parser(spec).items()))
         self.__insert = None
         self.__fieldwalker = None
 
@@ -186,7 +186,7 @@ class Updator(object):
 
 def parse_inc(field, value, array_filters):
     if not is_numeric_type(value):
-        val_repr_ = "{!r}" if isinstance(value, string_type) else "{}"
+        val_repr_ = "{!r}" if isinstance(value, string_types) else "{}"
         val_repr_ = val_repr_.format(value)
         msg = ("Cannot increment with non-numeric argument: "
                "{{{0}: {1}}}".format(field, val_repr_))
@@ -194,7 +194,7 @@ def parse_inc(field, value, array_filters):
 
     def _inc(fieldwalker):
 
-        def inc(node, inc_val):
+        def evaluator(node, inc_val):
             old_val = node.value
             if node.exists and not is_numeric_type(old_val):
                 _id = fieldwalker.doc["_id"]
@@ -217,14 +217,16 @@ def parse_inc(field, value, array_filters):
             else:
                 return (old_val or 0) + inc_val
 
-        _update(fieldwalker, field, value, inc, array_filters)
+        _update(fieldwalker, field, value, evaluator, array_filters)
 
     return _inc
 
 
 def parse_min(field, value, array_filters):
+
     def _min(fieldwalker):
-        def min(node, min_val):
+
+        def evaluator(node, min_val):
             old_val = node.value
             if node.exists:
                 old_val = Weighted(old_val)
@@ -233,14 +235,16 @@ def parse_min(field, value, array_filters):
             else:
                 return min_val
 
-        _update(fieldwalker, field, value, min, array_filters)
+        _update(fieldwalker, field, value, evaluator, array_filters)
 
     return _min
 
 
 def parse_max(field, value, array_filters):
+
     def _max(fieldwalker):
-        def max(node, max_val):
+
+        def evaluator(node, max_val):
             old_val = node.value
             if node.exists:
                 old_val = Weighted(old_val)
@@ -249,21 +253,22 @@ def parse_max(field, value, array_filters):
             else:
                 return max_val
 
-        _update(fieldwalker, field, value, max, array_filters)
+        _update(fieldwalker, field, value, evaluator, array_filters)
 
     return _max
 
 
 def parse_mul(field, value, array_filters):
     if not is_numeric_type(value):
-        val_repr_ = "{!r}" if isinstance(value, string_type) else "{}"
+        val_repr_ = "{!r}" if isinstance(value, string_types) else "{}"
         val_repr_ = val_repr_.format(value)
         msg = ("Cannot multiply with non-numeric argument: "
                "{{{0}: {1}}}".format(field, val_repr_))
         raise WriteError(msg, code=14)
 
     def _mul(fieldwalker):
-        def mul(node, mul_val):
+
+        def evaluator(node, mul_val):
             old_val = node.value
             if node.exists and not is_numeric_type(old_val):
                 _id = fieldwalker.doc["_id"]
@@ -286,7 +291,7 @@ def parse_mul(field, value, array_filters):
             else:
                 return (old_val or 0.0) * mul_val
 
-        _update(fieldwalker, field, value, mul, array_filters)
+        _update(fieldwalker, field, value, evaluator, array_filters)
 
     return _mul
 
@@ -298,7 +303,7 @@ def _get_array_member(fieldvalues):
 
 
 def parse_rename(field, new_field, array_filters):
-    if not isinstance(new_field, string_type):
+    if not isinstance(new_field, string_types):
         msg = ("The 'to' field for $rename must be a string: {0}: {1}"
                "".format(field, new_field))
         raise WriteError(msg, code=2)
@@ -320,7 +325,7 @@ def parse_rename(field, new_field, array_filters):
         probe.go(field).get()
         fieldvalues = probe.value
 
-        if not fieldvalues.exists:
+        if not fieldvalues.is_exists():
             return
 
         value = next(fieldvalues.iter_plain())
@@ -353,14 +358,18 @@ def parse_rename(field, new_field, array_filters):
 
 
 def parse_set(field, value, array_filters):
+
     def _set(fieldwalker):
+
         _update(fieldwalker, field, value, None, array_filters)
 
     return _set
 
 
 def parse_unset(field, _, array_filters):
+
     def _unset(fieldwalker):
+
         _drop(fieldwalker, field, array_filters)
 
     return _unset
@@ -399,9 +408,18 @@ def parse_currentDate(field, value, array_filters):
     return _currentDate
 
 
-def parse_add_to_set(field, value, array_filters):
+def parse_add_to_set(field, value_or_each, array_filters):
+    if (is_duckument_type(value_or_each) and
+            next(iter(value_or_each)) == "$each"):
+        value = EachAdder(value_or_each)
+        run_each = True
+    else:
+        value = value_or_each
+        run_each = False
+
     def _add_to_set(fieldwalker):
-        def add_to_set(node, new_elem):
+
+        def evaluator(node, new_elem):
             old_val = node.value
             if node.exists and not isinstance(old_val, list):
                 value_type = type(old_val).__name__
@@ -410,12 +428,17 @@ def parse_add_to_set(field, value, array_filters):
                        "".format(str(node), value_type))
                 raise WriteError(msg, code=2)
 
-            new_array = (old_val or [])[:]
-            if new_elem not in new_array:
-                new_array.append(new_elem)
+            if run_each:
+                eacher = new_elem
+                new_array = eacher(old_val)
+            else:
+                new_array = (old_val or [])[:]
+                if new_elem not in new_array:
+                    new_array.append(new_elem)
+
             return new_array
 
-        _update(fieldwalker, field, value, add_to_set, array_filters)
+        _update(fieldwalker, field, value, evaluator, array_filters)
 
     return _add_to_set
 
@@ -436,7 +459,8 @@ def parse_pop(field, value, array_filters):
             raise WriteError(msg_raw.format(field, value), code=9)
 
     def _pop(fieldwalker):
-        def pop(node, pop_ind):
+
+        def evaluator(node, pop_ind):
             old_val = node.value
             if node.exists and not isinstance(old_val, list):
                 value_type = type(old_val).__name__
@@ -453,7 +477,7 @@ def parse_pop(field, value, array_filters):
             else:
                 return old_val[1:]
 
-        _update(fieldwalker, field, value, pop, array_filters)
+        _update(fieldwalker, field, value, evaluator, array_filters)
 
     return _pop
 
@@ -471,7 +495,8 @@ def parse_pull(field, value_or_conditions, array_filters):
         queryfilter = QueryFilter({field: value_or_conditions})
 
     def _pull(fieldwalker):
-        def pull(node, _):
+
+        def evaluator(node, _):
             old_val = node.value
             if node.exists and not isinstance(old_val, list):
                 msg = "Cannot apply $pull to a non-array value"
@@ -489,14 +514,22 @@ def parse_pull(field, value_or_conditions, array_filters):
                     new_array.append(elem)
             return new_array
 
-        _update(fieldwalker, field, None, pull, array_filters)
+        _update(fieldwalker, field, None, evaluator, array_filters)
 
     return _pull
 
 
-def parse_push(field, value, array_filters):
+def parse_push(field, value_or_each, array_filters):
+    if is_duckument_type(value_or_each) and "$each" in value_or_each:
+        value = EachPusher(value_or_each)
+        run_each = True
+    else:
+        value = value_or_each
+        run_each = False
+
     def _push(fieldwalker):
-        def push(node, new_elem):
+
+        def evaluator(node, new_elem):
             old_val = node.value
             if node.exists and not isinstance(old_val, list):
                 value_type = type(old_val).__name__
@@ -506,11 +539,16 @@ def parse_push(field, value, array_filters):
                        "".format(str(node), value_type, _id))
                 raise WriteError(msg, code=2)
 
-            new_array = (old_val or [])[:]
-            new_array.append(new_elem)
+            if run_each:
+                eacher = new_elem
+                new_array = eacher(old_val)
+            else:
+                new_array = (old_val or [])[:]
+                new_array.append(new_elem)
+
             return new_array
 
-        _update(fieldwalker, field, value, push, array_filters)
+        _update(fieldwalker, field, value, evaluator, array_filters)
 
     return _push
 
@@ -523,7 +561,8 @@ def parse_pull_all(field, value, array_filters):
         raise WriteError(msg, code=2)
 
     def _pull_all(fieldwalker):
-        def pull_all(node, pull_list):
+
+        def evaluator(node, pull_list):
             old_val = node.value
             if node.exists and not isinstance(old_val, list):
                 msg = "Cannot apply $pull to a non-array value"
@@ -546,6 +585,166 @@ def parse_pull_all(field, value, array_filters):
             new_array = [elem for elem in old_val if elem not in pull_list]
             return new_array
 
-        _update(fieldwalker, field, value, pull_all, array_filters)
+        _update(fieldwalker, field, value, evaluator, array_filters)
 
     return _pull_all
+
+
+class EachAdder(object):
+
+    def __init__(self, spec):
+        spec = spec.copy()
+
+        self.mods = {
+            "$each": None,
+        }
+
+        for mod, value in spec.items():
+            try:
+                type_check = self.validators[mod]
+            except KeyError:
+                raise WriteError("Found unexpected fields after $each in "
+                                 "$addToSet: %s" % spec,
+                                 code=2)
+
+            self.mods[mod] = type_check(self, value)
+
+    def __call__(self, array):
+        new_array = (array or [])[:]
+        new_elems = self.mods["$each"][:]
+
+        new_array[0:] += [e for e in new_elems if e not in new_array]
+        return new_array
+
+    def _validate_each(self, each):
+        try:
+            each[:]
+        except TypeError:
+            type_name = type(each).__name__
+            raise WriteError("The argument to $each in $addToSet must be an "
+                             "array but it was of type %s" % type_name,
+                             code=14)
+        return each
+
+    validators = {
+        "$each": _validate_each,
+    }
+
+
+class EachPusher(object):
+
+    def __init__(self, spec):
+        spec = spec.copy()
+
+        self.mods = {
+            "$each": None,
+            "$position": None,
+            "$slice": None,
+            "$sort": None,
+        }
+
+        for mod, value in spec.items():
+            try:
+                type_check = self.validators[mod]
+            except KeyError:
+                raise WriteError("Unrecognized clause in $push: %s" % mod,
+                                 code=2)
+
+            self.mods[mod] = type_check(self, value)
+
+    def __call__(self, array):
+        new_array = (array or [])[:]
+        new_elems = self.mods["$each"][:]
+
+        position = self.mods["$position"]
+        slice = self.mods["$slice"]
+        sort = self.mods["$sort"]
+
+        if position is None:
+            new_array += new_elems
+        else:
+            new_array[:position] += new_elems
+
+        if slice is not None:
+            if slice >= 0:
+                new_array = new_array[:slice]
+            else:
+                new_array = new_array[slice:]
+
+        if sort is not None:
+            if is_duckument_type(sort):
+                fieldwalkers = list()
+                unsortable = list()
+                for elem in new_array:
+                    if is_duckument_type(elem):
+                        fieldwalkers.append(FieldWalker(elem))
+                    else:
+                        unsortable.append(elem)
+
+                ordered = ordering(fieldwalkers, sort)
+                new_array = [f.doc for f in ordered]
+
+                if unsortable:
+                    is_reverse = bool(1 - next(iter(sort.values())))
+                    if is_reverse:
+                        new_array += unsortable
+                    else:
+                        new_array[:0] += unsortable
+
+            else:
+                is_reverse = bool(1 - sort)
+                ordered = sorted((Weighted(e) for e in new_array),
+                                 reverse=is_reverse)
+                new_array = [w.value for w in ordered]
+
+        return new_array
+
+    def _validate_each(self, each):
+        try:
+            each[:]
+        except TypeError:
+            type_name = type(each).__name__
+            raise WriteError("The argument to $each in $push must be an "
+                             "array but it was of type: %s" % type_name,
+                             code=2)
+        return each
+
+    def _validate_position(self, position):
+        if not is_integer_type(position):
+            type_name = type(position).__name__
+            raise WriteError("The value for $position must be an integer "
+                             "value, not of type: %s" % type_name,
+                             code=2)
+        return position
+
+    def _validate_slice(self, slice):
+        if not is_integer_type(slice):
+            type_name = type(slice).__name__
+            raise WriteError("The value for $slice must be an integer "
+                             "value but was given type: %s" % type_name,
+                             code=2)
+        return slice
+
+    def _validate_sort(self, sort, int_only=False):
+        if is_integer_type(sort) or int_only:
+            if sort not in (1, -1):
+                raise WriteError("The $sort element value must be either "
+                                 "1 or -1",
+                                 code=2)
+            return sort
+
+        if is_duckument_type(sort):
+            for key, value in sort.items():
+                self._validate_sort(value, int_only=True)
+            return sort
+
+        raise WriteError("The $sort is invalid: use 1/-1 to sort the whole "
+                         "element, or {field:1/-1} to sort embedded fields",
+                         code=2)
+
+    validators = {
+        "$each": _validate_each,
+        "$position": _validate_position,
+        "$slice": _validate_slice,
+        "$sort": _validate_sort,
+    }
